/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II
======================================================================*/

#include "quakedef.h"
#include "renderer\glquake.h"
#include "server\world.h"
#include "client\client.h"
#include "client\chase.h"

namespace Axiom
{
	namespace Client
	{
		Chase chase;

		cvar_t	chase_back		("chase_back",		"100");
		cvar_t	chase_up		("chase_up",		"16");
		cvar_t	chase_right		("chase_right",		"0");
		cvar_t	chase_active	("chase_active",	"0");

		/*================================================================================
			TODO: impact on bmodels, monsters
		================================================================================*/
		void Chase::TraceLine (vec3_t start, vec3_t end, vec3_t impact)
		{
			trace_t	trace;

			memset (&trace, 0, sizeof(trace));
			SV_RecursiveHullCheck(cl.worldmodel->hulls, 0, 0, 1, start, end, &trace);

			VectorCopy (trace.endpos, impact);
		}

		/*================================================================================
			Orient client based on camera. Called after input
		================================================================================*/
		void Chase::Chase_UpdateForClient()
		{
			// place camera
			// assign client angles to camera
			// see where camera points
			// adjust client angles to point at the same place
		}

		/*================================================================================
			Orient camera based on client. Called before drawing.
			TODO: stay at least 8 units away from all walls in this leaf.
		================================================================================*/
		void Chase::Chase_UpdateForDrawing()
		{
			int		i;
			vec3_t	forward, up, right;
			vec3_t	ideal, crosshair, temp;
		
			AngleVectors(cl.viewangles, forward, right, up);
		
			// calc ideal camera location before checking for walls
			for(i = 0; i < 3; i++)
				ideal[i] = cl.viewent.origin[i] - forward[i] * chase_back.value + right[i] * chase_right.value;
				//+ up[i]*chase_up.value;
		
			ideal[2] = cl.viewent.origin[2] + chase_up.value;
		
			// make sure camera is not in or behind a wall
			TraceLine(r_refdef.vieworg, ideal, temp);

			if(Length(temp) != 0)
				VectorCopy(temp, ideal);
		
			// place camera
			VectorCopy(ideal, r_refdef.vieworg);

			// find the spot the player is looking at
			VectorMA	(cl.viewent.origin, 4096, forward, temp);
			TraceLine	(cl.viewent.origin, temp, crosshair);

			// calculate camera angles to look at the same spot
			VectorSubtract	(crosshair,	r_refdef.vieworg, temp);
			VectorAngles	(temp,		r_refdef.viewangles);

			if(r_refdef.viewangles[PITCH] == 90 || r_refdef.viewangles[PITCH] == -90)
				r_refdef.viewangles[YAW] = cl.viewangles[YAW];
		}
	}
}