/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II

	Notes:
	Player eye positioning

	The view is allowed to move slightly from it's true position for bobbing,
	but if it exceeds 8 pixels linear distance (spherical, not box), the list of
	entities sent from the server may not include everything in the pvs, especially
	when crossing a water boudnary.
======================================================================*/

#include <math.h>

#include "quakedef.h"
#include "cmd.h"
#include "console.h"
#include "host.h"

#include "client\chase.h"
#include "client\cl_input.h"

namespace Axiom
{
	namespace Client
	{
		cvar_t	scr_ofsx		("scr_ofsx",	"0");
		cvar_t	scr_ofsy		("scr_ofsy",	"0");
		cvar_t	scr_ofsz		("scr_ofsz",	"0");

		cvar_t	cl_rollspeed	("cl_rollspeed",	"200");
		cvar_t	cl_rollangle	("cl_rollangle",	"1");

		cvar_t	cl_bob			("cl_bob",		"0");
		cvar_t	cl_bobcycle		("cl_bobcycle",	"1");
		cvar_t	cl_bobup		("cl_bobup",	"0");

		cvar_t	v_kicktime		("v_kicktime",	"0.5");
		cvar_t	v_kickroll		("v_kickroll",	"0.6");
		cvar_t	v_kickpitch		("v_kickpitch",	"0.6");
		cvar_t	v_gunkick		("v_gunkick",	"2");

		cvar_t	v_iyaw_cycle	("v_iyaw_cycle",	"2");
		cvar_t	v_iroll_cycle	("v_iroll_cycle",	"2");
		cvar_t	v_ipitch_cycle	("v_ipitch_cycle",	"2");
		cvar_t	v_iyaw_level	("v_iyaw_level",	"0.3");
		cvar_t	v_iroll_level	("v_iroll_level",	"0.1");
		cvar_t	v_ipitch_level	("v_ipitch_level",	"0.3");

		cvar_t	v_idlescale		("v_idlescale",	"0");
		cvar_t	crosshair		("crosshair",	"1",CVAR_ARCHIVE);

		cvar_t r_viewmodeloffset	("r_viewmodeloffset",	"0",CVAR_ARCHIVE);
		cvar_t gl_cshiftpercent		("gl_cshiftpercent",	"100");

		float		v_dmg_time, v_dmg_roll, v_dmg_pitch;
		extern int	in_forward, in_forward2, in_back;

		// Copied from cl.punchangle.  0 is current, 1 is previous value.
		// Never the same unless map just loaded.
		vec3_t	v_punchangles[2];

		/*================================================================================
			Used by view and sv_user
		================================================================================*/
		vec3_t	forward, right, up;

		float V_CalcRoll(vec3_t angles, vec3_t velocity)
		{
			float	sign;
			float	side;
			float	value;

			AngleVectors (angles, forward, right, up);
			side = DotProduct (velocity, right);
			sign = side < 0 ? -1 : 1;
			side = fabs(side);

			value = cl_rollangle.value;
			//if (cl.inwater)
				//value *= 6;

			if (side < cl_rollspeed.value)
				side = side * value / cl_rollspeed.value;
			else
				side = value;

			return side*sign;

		}

		/*================================================================================
			CalcBob
		================================================================================*/
		float V_CalcBob()
		{
			float	bob;
			float	cycle;

			cycle = cl.time - (int)(cl.time/cl_bobcycle.value)*cl_bobcycle.value;
			cycle /= cl_bobcycle.value;
			if (cycle < cl_bobup.value)
				cycle = M_PI * cycle / cl_bobup.value;
			else
				cycle = M_PI + M_PI*(cycle-cl_bobup.value)/(1.0 - cl_bobup.value);

			// bob is proportional to velocity in the xy plane
			// (don't count Z, or jumping messes it up)

			bob = sqrt(cl.velocity[0]*cl.velocity[0] + cl.velocity[1]*cl.velocity[1]) * cl_bob.value;
			//Con_Printf ("speed: %5.1f\n", Length(cl.velocity));
			bob = bob*0.3 + bob*0.7*sin(cycle);
			if (bob > 4)
				bob = 4;
			else if (bob < -7)
				bob = -7;
			return bob;

		}

		cvar_t	v_centermove	("v_centermove",	"0.15");
		cvar_t	v_centerspeed	("v_centerspeed",	"500");


		/*================================================================================
			StartPitchDrift
		================================================================================*/
		void V_StartPitchDrift()
		{
		#if 1
			if (cl.laststop == cl.time)
			{
				return;		// something else is keeping it from drifting
			}
		#endif
			if (cl.nodrift || !cl.pitchvel)
			{
				cl.pitchvel = v_centerspeed.value;
				cl.nodrift = false;
				cl.driftmove = 0;
			}
		}

		void V_StopPitchDrift (void)
		{
			cl.laststop = cl.time;
			cl.nodrift = true;
			cl.pitchvel = 0;
		}

		/*================================================================================
			Moves the client pitch angle towards cl.idealpitch sent by the server.

			If the user is adjusting pitch manually, either with lookup/lookdown,
			mlook and mouse, or klook and keyboard, pitch drifting is constantly stopped.

			Drifting is enabled when the center view key is hit, mlook is released and
			lookspring is non 0, or when
		================================================================================*/
		void V_DriftPitch()
		{
			float		delta, move;

			if (noclip_anglehack || !cl.onground || cls.demoplayback )
			//FIXME: noclip_anglehack is set on the server, so in a nonlocal game this won't work.
			{
				cl.driftmove = 0;
				cl.pitchvel = 0;
				return;
			}
	
			// don't count small mouse motion
			if (cl.nodrift)
			{
				if ( fabs(cl.cmd.forwardmove) < cl_forwardspeed.value)
					cl.driftmove = 0;
				else
					cl.driftmove += host_frametime;

				if ( cl.driftmove > v_centermove.value)
				{
					V_StartPitchDrift ();
				}
				return;
			}

			delta = cl.idealpitch - cl.viewangles[PITCH];

			if (!delta)
			{
				cl.pitchvel = 0;
				return;
			}

			move = host_frametime * cl.pitchvel;
			cl.pitchvel += host_frametime * v_centerspeed.value;

			//Con_Printf ("move: %f (%f)\n", move, host_frametime);

			if (delta > 0)
			{
				if (move > delta)
				{
					cl.pitchvel = 0;
					move = delta;
				}
				cl.viewangles[PITCH] += move;
			}
			else if (delta < 0)
			{
				if (move > -delta)
				{
					cl.pitchvel = 0;
					move = -delta;
				}
				cl.viewangles[PITCH] -= move;
			}
		}

		cshift_t	cshift_empty = { {130,80,50}, 0 };
		cshift_t	cshift_water = { {130,80,50}, 128 };
		cshift_t	cshift_slime = { {0,25,5}, 150 };
		cshift_t	cshift_lava = { {255,80,0}, 150 };

		float		v_blend[4];		// rgba 0.0 - 1.0

		// Deleted BuildGammaTable(), V_CheckGamma(), gammatable[], and ramps[][]

		/*================================================================================
			ParseDamage
		================================================================================*/
		void V_ParseDamage (void)
		{
			int		armor, blood;
			vec3_t	from;
			int		i;
			vec3_t	forward, right, up;
			entity_t	*ent;
			float	side;
			float	count;

			armor = MSG_ReadByte ();
			blood = MSG_ReadByte ();
			for (i=0 ; i<3 ; i++)
				from[i] = MSG_ReadCoord ();

			count = blood*0.5 + armor*0.5;
			if (count < 10)
				count = 10;

			cl.faceanimtime = cl.time + 0.2;		// but sbar face into pain frame

			cl.cshifts[CSHIFT_DAMAGE].percent += 3*count;
			if (cl.cshifts[CSHIFT_DAMAGE].percent < 0)
				cl.cshifts[CSHIFT_DAMAGE].percent = 0;
			if (cl.cshifts[CSHIFT_DAMAGE].percent > 150)
				cl.cshifts[CSHIFT_DAMAGE].percent = 150;

			if (armor > blood)
			{
				cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 200;
				cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 100;
				cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 100;
			}
			else if (armor)
			{
				cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 220;
				cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 50;
				cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 50;
			}
			else
			{
				cl.cshifts[CSHIFT_DAMAGE].destcolor[0] = 255;
				cl.cshifts[CSHIFT_DAMAGE].destcolor[1] = 0;
				cl.cshifts[CSHIFT_DAMAGE].destcolor[2] = 0;
			}

			//
			// calculate view angle kicks
			//
			ent = &cl_entities[cl.viewentity];

			VectorSubtract (from, ent->origin, from);
			VectorNormalize (from);

			AngleVectors (ent->angles, forward, right, up);

			side = DotProduct (from, right);
			v_dmg_roll = count*side*v_kickroll.value;

			side = DotProduct (from, forward);
			v_dmg_pitch = count*side*v_kickpitch.value;

			v_dmg_time = v_kicktime.value;
		}

		/*================================================================================
			Cshift
		================================================================================*/
		void V_cshift_f()
		{
			cshift_empty.destcolor[0] = atoi(Cmd_Argv(1));
			cshift_empty.destcolor[1] = atoi(Cmd_Argv(2));
			cshift_empty.destcolor[2] = atoi(Cmd_Argv(3));
			cshift_empty.percent = atoi(Cmd_Argv(4));
		}

		/*================================================================================
			When you run over an item, the server sends this command
		================================================================================*/
		void V_BonusFlash_f()
		{
			cl.cshifts[CSHIFT_BONUS].destcolor[0] = 215;
			cl.cshifts[CSHIFT_BONUS].destcolor[1] = 186;
			cl.cshifts[CSHIFT_BONUS].destcolor[2] = 69;
			cl.cshifts[CSHIFT_BONUS].percent = 50;
		}

		/*================================================================================
			Underwater, lava, etc each has a color shift.
		================================================================================*/
		void V_SetContentsColor(int contents)
		{
			switch (contents)
			{
			case CONTENTS_EMPTY:
			case CONTENTS_SOLID:
			case CONTENTS_SKY: //johnfitz -- no blend in sky
				cl.cshifts[CSHIFT_CONTENTS] = cshift_empty;
				break;
			case CONTENTS_LAVA:
				cl.cshifts[CSHIFT_CONTENTS] = cshift_lava;
				break;
			case CONTENTS_SLIME:
				cl.cshifts[CSHIFT_CONTENTS] = cshift_slime;
				break;
			default:
				cl.cshifts[CSHIFT_CONTENTS] = cshift_water;
			}
		}

		/*================================================================================
			CalcPowerupCshift
		================================================================================*/
		void V_CalcPowerupCshift (void)
		{
			if (cl.items & IT_QUAD)
			{
				cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 0;
				cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 0;
				cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 255;
				cl.cshifts[CSHIFT_POWERUP].percent = 30;
			}
			else if (cl.items & IT_SUIT)
			{
				cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 0;
				cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 255;
				cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 0;
				cl.cshifts[CSHIFT_POWERUP].percent = 20;
			}
			else if (cl.items & IT_INVISIBILITY)
			{
				cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 100;
				cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 100;
				cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 100;
				cl.cshifts[CSHIFT_POWERUP].percent = 100;
			}
			else if (cl.items & IT_INVULNERABILITY)
			{
				cl.cshifts[CSHIFT_POWERUP].destcolor[0] = 255;
				cl.cshifts[CSHIFT_POWERUP].destcolor[1] = 255;
				cl.cshifts[CSHIFT_POWERUP].destcolor[2] = 0;
				cl.cshifts[CSHIFT_POWERUP].percent = 30;
			}
			else
				cl.cshifts[CSHIFT_POWERUP].percent = 0;
		}

		/*================================================================================
			CalcBlend
		================================================================================*/
		void V_CalcBlend (void)
		{
			float	r, g, b, a, a2;
			int		j;

			r = 0;
			g = 0;
			b = 0;
			a = 0;

			for (j=0 ; j<NUM_CSHIFTS ; j++)
			{
				if (!gl_cshiftpercent.value)
					continue;

				//johnfitz -- only apply leaf contents color shifts during intermission
				if (cl.intermission && j != CSHIFT_CONTENTS)
					continue;
				//johnfitz

				a2 = ((cl.cshifts[j].percent * gl_cshiftpercent.value) / 100.0) / 255.0;
				if (!a2)
					continue;
				a = a + a2*(1-a);
				a2 = a2/a;
				r = r*(1-a2) + cl.cshifts[j].destcolor[0]*a2;
				g = g*(1-a2) + cl.cshifts[j].destcolor[1]*a2;
				b = b*(1-a2) + cl.cshifts[j].destcolor[2]*a2;
			}

			v_blend[0] = r/255.0;
			v_blend[1] = g/255.0;
			v_blend[2] = b/255.0;
			v_blend[3] = a;
			if (v_blend[3] > 1)
				v_blend[3] = 1;
			if (v_blend[3] < 0)
				v_blend[3] = 0;
		}

		/*================================================================================
			UpdateBlend
		================================================================================*/
		void V_UpdateBlend()
		{
			int		i, j;
			bool	blend_changed;

			V_CalcPowerupCshift ();

			blend_changed = false;

			for (i=0 ; i<NUM_CSHIFTS ; i++)
			{
				if (cl.cshifts[i].percent != cl.prev_cshifts[i].percent)
				{
					blend_changed = true;
					cl.prev_cshifts[i].percent = cl.cshifts[i].percent;
				}
				for (j=0 ; j<3 ; j++)
					if (cl.cshifts[i].destcolor[j] != cl.prev_cshifts[i].destcolor[j])
					{
						blend_changed = true;
						cl.prev_cshifts[i].destcolor[j] = cl.cshifts[i].destcolor[j];
					}
			}
	
			// drop the damage value
			cl.cshifts[CSHIFT_DAMAGE].percent -= host_frametime*150;
			if (cl.cshifts[CSHIFT_DAMAGE].percent <= 0)
				cl.cshifts[CSHIFT_DAMAGE].percent = 0;
	
			// drop the bonus value
			cl.cshifts[CSHIFT_BONUS].percent -= host_frametime*100;
			if (cl.cshifts[CSHIFT_BONUS].percent <= 0)
				cl.cshifts[CSHIFT_BONUS].percent = 0;

			if (blend_changed)
				V_CalcBlend ();
		}

		/*================================================================================
			PolyBlend
		================================================================================*/
		void V_PolyBlend()
		{
			if (!gl_polyblend.value || !v_blend[3])
				return;

			GL_DisableMultitexture();

			glDisable (GL_ALPHA_TEST);
			glDisable (GL_TEXTURE_2D);
			glDisable (GL_DEPTH_TEST);
			glEnable (GL_BLEND);

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity ();
			glOrtho (0, 1, 1, 0, -99999, 99999);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity ();

			glColor4fv (v_blend);

			glBegin (GL_QUADS);
			glVertex2f (0,0);
			glVertex2f (1, 0);
			glVertex2f (1, 1);
			glVertex2f (0, 1);
			glEnd ();

			glDisable (GL_BLEND);
			glEnable (GL_DEPTH_TEST);
			glEnable (GL_TEXTURE_2D);
			glEnable (GL_ALPHA_TEST);
		}

		float angledelta(float a)
		{
			a = anglemod(a);
			if (a > 180)
				a -= 360;
			return a;
		}

		/*================================================================================
			CalcGunAngle
		================================================================================*/
		void CalcGunAngle (void)
		{
			float	yaw, pitch, move;
			static float oldyaw = 0;
			static float oldpitch = 0;

			yaw = r_refdef.viewangles[YAW];
			pitch = -r_refdef.viewangles[PITCH];

			yaw = angledelta(yaw - r_refdef.viewangles[YAW]) * 0.4;
			if (yaw > 10)
				yaw = 10;
			if (yaw < -10)
				yaw = -10;
			pitch = angledelta(-pitch - r_refdef.viewangles[PITCH]) * 0.4;
			if (pitch > 10)
				pitch = 10;
			if (pitch < -10)
				pitch = -10;
			move = host_frametime * 20;
			if (yaw > oldyaw)
			{
				if (oldyaw + move < yaw)
					yaw = oldyaw + move;
			}
			else
			{
				if (oldyaw - move > yaw)
					yaw = oldyaw - move;
			}

			if (pitch > oldpitch)
			{
				if (oldpitch + move < pitch)
					pitch = oldpitch + move;
			}
			else
			{
				if (oldpitch - move > pitch)
					pitch = oldpitch - move;
			}

			oldyaw = yaw;
			oldpitch = pitch;

			cl.viewent.angles[YAW] = r_refdef.viewangles[YAW] + yaw;
			cl.viewent.angles[PITCH] = - (r_refdef.viewangles[PITCH] + pitch);

			cl.viewent.angles[ROLL] -= v_idlescale.value * sin(cl.time*v_iroll_cycle.value) * v_iroll_level.value;
			cl.viewent.angles[PITCH] -= v_idlescale.value * sin(cl.time*v_ipitch_cycle.value) * v_ipitch_level.value;
			cl.viewent.angles[YAW] -= v_idlescale.value * sin(cl.time*v_iyaw_cycle.value) * v_iyaw_level.value;
		}

		/*================================================================================
			BoundOffsets
		================================================================================*/
		void V_BoundOffsets (void)
		{
			entity_t	*ent;

			ent = &cl_entities[cl.viewentity];

			// absolutely bound refresh reletive to entity clipping hull
			// so the view can never be inside a solid wall

			if (r_refdef.vieworg[0] < ent->origin[0] - 14)
				r_refdef.vieworg[0] = ent->origin[0] - 14;
			else if (r_refdef.vieworg[0] > ent->origin[0] + 14)
				r_refdef.vieworg[0] = ent->origin[0] + 14;
			if (r_refdef.vieworg[1] < ent->origin[1] - 14)
				r_refdef.vieworg[1] = ent->origin[1] - 14;
			else if (r_refdef.vieworg[1] > ent->origin[1] + 14)
				r_refdef.vieworg[1] = ent->origin[1] + 14;
			if (r_refdef.vieworg[2] < ent->origin[2] - 22)
				r_refdef.vieworg[2] = ent->origin[2] - 22;
			else if (r_refdef.vieworg[2] > ent->origin[2] + 30)
				r_refdef.vieworg[2] = ent->origin[2] + 30;
		}

		/*================================================================================
			Idle swaying
		================================================================================*/
		void V_AddIdle (void)
		{
			r_refdef.viewangles[ROLL]	+= v_idlescale.value * sin(cl.time*v_iroll_cycle.value) * v_iroll_level.value;
			r_refdef.viewangles[PITCH]	+= v_idlescale.value * sin(cl.time*v_ipitch_cycle.value) * v_ipitch_level.value;
			r_refdef.viewangles[YAW]	+= v_idlescale.value * sin(cl.time*v_iyaw_cycle.value) * v_iyaw_level.value;
		}

		/*================================================================================
			Roll is induced by movement and damage
		================================================================================*/
		void V_CalcViewRoll (void)
		{
			float		side;

			side = V_CalcRoll (cl_entities[cl.viewentity].angles, cl.velocity);
			r_refdef.viewangles[ROLL] += side;

			if (v_dmg_time > 0)
			{
				r_refdef.viewangles[ROLL] += v_dmg_time/v_kicktime.value*v_dmg_roll;
				r_refdef.viewangles[PITCH] += v_dmg_time/v_kicktime.value*v_dmg_pitch;
				v_dmg_time -= host_frametime;
			}

			if (cl.stats[STAT_HEALTH] <= 0)
			{
				r_refdef.viewangles[ROLL] = 80;	// dead view angle
				return;
			}

		}

		/*================================================================================
			CalcIntermissionRefDef
		================================================================================*/
		void V_CalcIntermissionRefdef()
		{
			entity_t	*ent, *view;
			float		old;

			// ent is the player model (visible when out of body)
			ent = &cl_entities[cl.viewentity];
			// view is the weapon model (only visible from inside body)
			view = &cl.viewent;

			VectorCopy (ent->origin, r_refdef.vieworg);
			VectorCopy (ent->angles, r_refdef.viewangles);
			view->model = NULL;

			// allways idle in intermission
			old = v_idlescale.value;
			v_idlescale.value = 1;
			V_AddIdle ();
			v_idlescale.value = old;
		}

		/*================================================================================
			CalcRefDef
		================================================================================*/
		void V_CalcRefdef()
		{
			entity_t	*ent, *view;
			int			i;
			vec3_t		forward, right, up;
			vec3_t		angles;
			float		bob;
			static float oldz = 0;
			static vec3_t punch = {0,0,0};	// v_gunkick
			float delta;					// v_gunkick

			V_DriftPitch ();

			// ent is the player model (visible when out of body)
			ent = &cl_entities[cl.viewentity];
			// view is the weapon model (only visible from inside body)
			view = &cl.viewent;

			// transform the view offset by the model's matrix to get the offset from
			// model origin for the view
			ent->angles[YAW]	= cl.viewangles[YAW];	// the model should face the view dir
			ent->angles[PITCH]	= -cl.viewangles[PITCH];	// the model should face the view dir

			bob = V_CalcBob ();

			// refresh position
			VectorCopy (ent->origin, r_refdef.vieworg);
			r_refdef.vieworg[2] += cl.viewheight + bob;

			// never let it sit exactly on a node line, because a water plane can
			// dissapear when viewed with the eye exactly on it.
			// the server protocol only specifies to 1/16 pixel, so add 1/32 in each axis
			r_refdef.vieworg[0] += 1.0/32;
			r_refdef.vieworg[1] += 1.0/32;
			r_refdef.vieworg[2] += 1.0/32;

			VectorCopy(cl.viewangles, r_refdef.viewangles);
			V_CalcViewRoll();
			V_AddIdle();

			// offsets
			angles[PITCH]	= -ent->angles[PITCH]; // because entity pitches are actually backward
			angles[YAW]		= ent->angles[YAW];
			angles[ROLL]	= ent->angles[ROLL];

			AngleVectors (angles, forward, right, up);

			if(cl.maxclients <= 1) //johnfitz -- moved cheat-protection here from V_RenderView
				for(i = 0; i < 3; i++)
					r_refdef.vieworg[i] += scr_ofsx.value * forward[i] + scr_ofsy.value * right[i] + scr_ofsz.value * up[i];

			V_BoundOffsets();

			// set up gun position
			VectorCopy(cl.viewangles, view->angles);

			CalcGunAngle();

			VectorCopy(ent->origin, view->origin);
			view->origin[2] += cl.viewheight;

		#if 1
			for(i = 0; i < 3; i++)
			{
				view->origin[i] += forward[i];
				view->origin[i] += right[i];
				view->origin[i] += up[i];
			}
			view->origin[2] += bob;
		#endif

		#if 0
			VectorCopy(r_refdef.vieworg, view->origin);
			VectorMA(view->origin, bob * 0.4, forward, view->origin);
	
			if(r_viewmodeloffset.string[0])
			{
				float offset[3];
				int size = sizeof(offset) / sizeof(offset[0]);
		
				ParseFloats(r_viewmodeloffset.string, offset, &size);
				VectorMA(view->origin,  offset[0], right,   view->origin);
				VectorMA(view->origin, -offset[1], up,      view->origin);
				VectorMA(view->origin,  offset[2], forward, view->origin);
			}
		#endif

			// fudge position around to keep amount of weapon visible
			// roughly equal with different FOV

			view->model		= cl.model_precache[cl.stats[STAT_WEAPON]];
			view->frame		= cl.stats[STAT_WEAPONFRAME];
			view->colormap	= vid.colormap;

			// johnfitz -- v_gunkick
			if (v_gunkick.value == 1) //original quake kick
				VectorAdd (r_refdef.viewangles, cl.punchangle, r_refdef.viewangles);
			if (v_gunkick.value == 2) //lerped kick
			{
				for (i=0; i<3; i++)
					if (punch[i] != v_punchangles[0][i])
					{
						//speed determined by how far we need to lerp in 1/10th of a second
						delta = (v_punchangles[0][i]-v_punchangles[1][i]) * host_frametime * 10;

						if (delta > 0)
							punch[i] = min(punch[i]+delta, v_punchangles[0][i]);
						else if (delta < 0)
							punch[i] = max(punch[i]+delta, v_punchangles[0][i]);
					}

				VectorAdd (r_refdef.viewangles, punch, r_refdef.viewangles);
			}
	
			// smooth out stair step ups
			if (!noclip_anglehack && cl.onground && ent->origin[2] - oldz > 0) //johnfitz -- added exception for noclip
			//FIXME: noclip_anglehack is set on the server, so in a nonlocal game this won't work.
			{
				float steptime;

				steptime = cl.time - cl.oldtime;
				if (steptime < 0)
					//FIXME	I_Error ("steptime < 0");
					steptime = 0;

				oldz += steptime * 80;
				if (oldz > ent->origin[2])
					oldz = ent->origin[2];
				if (ent->origin[2] - oldz > 12)
					oldz = ent->origin[2] - 12;
				r_refdef.vieworg[2] += oldz - ent->origin[2];
				view->origin[2] += oldz - ent->origin[2];
			}
			else
				oldz = ent->origin[2];

			if (chase_active.value)
				chase.Chase_UpdateForDrawing ();
		}

		/*================================================================================
			Th player's clipping box goes from (-16 -16 -24) to (16 16 32) from
			the entity origin, so any view position inside that will be valid
		================================================================================*/
		extern vrect_t	scr_vrect;

		void V_RenderView()
		{
			if (con_forcedup)
				return;

			if (cl.intermission)
				V_CalcIntermissionRefdef ();
			else if (!cl.paused /* && (cl.maxclients > 1 || key_dest == key_game) */)
				V_CalcRefdef ();

			//johnfitz -- removed lcd code

			R_RenderView();

			V_PolyBlend(); //johnfitz -- moved here from R_Renderview ();
		}

		/*================================================================================
			Init
		================================================================================*/
		void V_Init (void)
		{
		}
		
		cmd_t V_cshift_f_Cmd			("v_cshift",	V_cshift_f);
		cmd_t V_BonusFlash_f_Cmd		("bf",			V_BonusFlash_f);
		cmd_t V_StartPitchDrift_f_Cmd	("centerview",	V_StartPitchDrift);
	}
}