/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II
======================================================================*/

/*
Any number of commands can be added in a frame, from several different sources.
Most commands come from either keybindings or console line input, but remote
servers can also send across commands and entire text files can be execed.

The + command line options are also added to the command buffer.

The game starts with a Cbuf_AddText ("exec quake.rc\n"); Cbuf_Execute ();
*/

#ifndef CMD_H
#define CMD_H

#include "cvar.h"
#include "mathlib.h"

#define	MAX_ALIAS_NAME	32
#define CMDLINE_LENGTH 256

#define CMD_NORMAL		0
#define CMD_COMPAT		1

typedef void (*xcommand_t) (void);

typedef enum
{
	src_client,		// came in over a net connection as a clc_stringcmd
					// host_client will be valid during this state.
	src_command		// from the command buffer
} cmd_source_t;

extern	cmd_source_t	cmd_source;

namespace Axiom
{
	class cmd_t
	{
	public:
		cmd_t(char *cmdname, xcommand_t cmdcmd);

		char			*name;
		int				usage;
		xcommand_t		function;
		cmd_t *next;
	};
}

typedef struct cmdalias_s
{
	struct cmdalias_s	*next;
	char	name[MAX_ALIAS_NAME];
	char	*value;
} cmdalias_t;

class complist_t
{
public:
	// note - !!!if the order here is changed then the struct inits below also need to be changed!!!
	// search for every occurance of bsearch and do the necessary...
	char *name;
	cmdalias_t *als;
	cmd_t *cmd;
	cvar_t *var;
};

namespace Axiom
{
	void Cmd_Wait_f();
	void Cbuf_Init();
	void Cbuf_AddText(char *text);
	void Cbuf_InsertText(char *text);
	void Cbuf_Execute();
	void Cmd_StuffCmds_f();
	void Cmd_Exec_f();
	void Cmd_Echo_f();
	char *CopyString(char *in);
	void Cmd_Alias_f();
	void Cmd_Unalias_f();
	void Cmd_Unaliasall_f();
	void Cmd_List_f();
	void Cmd_Init();
	int Cmd_Argc();
	char *Cmd_Argv(int arg);
	char *Cmd_Args();
	void Cmd_TokenizeString(char *text);
	//void Cmd_AddCommand(char *cmd_name, xcommand_t function);
	bool Cmd_Exists(char *cmd_name);
	char *Cmd_CompleteCommand(char *partial);
	void Cmd_ExecuteString(char *text, cmd_source_t src);
	void Cmd_ForwardToServer();
	int Cmd_CheckParm(char *parm);

	int CmdCvarCompareFunc(const void *a, const void *b);
	void Cmd_BuildCompletionList();
}

#endif 