/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake(props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II
======================================================================*/

#ifndef COMMON_H
#define COMMON_H

#include <stdio.h>

#include "zone.h"
#include "bspfile.h"

#define	QUAKE_GAME
#define	GAMENAME	"id1"

#define	VERSION				1.09
#define	FITZQUAKE_VERSION	0.85

//define	PARANOID			// speed sapping error checking

#if !defined BYTE_DEFINED
typedef unsigned char 		byte;
#define BYTE_DEFINED 1
#endif

#define MAX_NUM_ARGVS	50

// up / down
#define	PITCH	0

// left / right
#define	YAW		1

// fall over
#define	ROLL	2

#define	MAX_QPATH		64			// max length of a quake game pathname
#define	MAX_OSPATH		128			// max length of a filesystem pathname

#define	ON_EPSILON		0.1			// point on plane side epsilon

#define	MAX_MSGLEN			32000	// max length of a reliable message //johnfitz -- was 8000
#define	MAX_DATAGRAM		32000	// max length of unreliable message //johnfitz -- was 1024

#define DATAGRAM_MTU		1400	// johnfitz -- actual limit for unreliable messages to nonlocal clients

// per-level limits
#define MIN_EDICTS		256			// johnfitz -- lowest allowed value for max_edicts cvar
#define MAX_EDICTS		32000		// johnfitz -- highest allowed value for max_edicts cvar
									// ents past 8192 can't play sounds in the standard protocol
#define	MAX_LIGHTSTYLES	64
#define	MAX_MODELS		2048		// johnfitz -- was 256
#define	MAX_SOUNDS		1024		// johnfitz -- was 256

#define	SAVEGAME_COMMENT_LENGTH	39

#define	MAX_STYLESTRING	64

typedef struct sizebuf_s
{
	bool	allowoverflow;	// if false, do a Sys_Error
	bool	overflowed;		// set to true if the buffer size failed
	byte	*data;
	int		maxsize;
	int		cursize;
} sizebuf_t;

typedef struct link_s
{
	struct link_s	*prev, *next;
} link_t;

//(type *)STRUCT_FROM_LINK(link_t *link, type, member)
// ent = STRUCT_FROM_LINK(link,entity_t,order)
// FIXME: remove this mess!
#define	STRUCT_FROM_LINK(l,t,m)((t *)((byte *)l -(int)&(((t *)0)->m)))

#ifndef NULL
#define NULL((void *)0)
#endif

#define Q_MAXCHAR	((char)0x7f)
#define Q_MAXSHORT	((short)0x7fff)
#define Q_MAXINT	((int)0x7fffffff)
#define Q_MAXLONG	((int)0x7fffffff)
#define Q_MAXFLOAT	((int)0x7fffffff)

#define Q_MINCHAR	((char)0x80)
#define Q_MINSHORT	((short)0x8000)
#define Q_MININT 	((int)0x80000000)
#define Q_MINLONG	((int)0x80000000)
#define Q_MINFLOAT	((int)0x7fffffff)

extern	bool		bigendien;

// in memory
typedef struct
{
	char    name[MAX_QPATH];
	int             filepos, filelen;
} packfile_t;

typedef struct pack_s
{
	char		filename[MAX_OSPATH];
	int			handle;
	FILE		*pak;
	int			numfiles;
	packfile_t	*files;
} pack_t;

// on disk
typedef struct
{
	char	name[56];
	int		filepos, filelen;
} dpackfile_t;

typedef struct
{
	char    id[4];
	int             dirofs;
	int             dirlen;
} dpackheader_t;

namespace Axiom
{
	void SZ_Alloc(sizebuf_t *buf, int startsize);
	void SZ_Free(sizebuf_t *buf);
	void SZ_Clear(sizebuf_t *buf);
	void *SZ_GetSpace(sizebuf_t *buf, int length);
	void SZ_Write(sizebuf_t *buf, void *data, int length);
	void SZ_Print(sizebuf_t *buf, char *data);	// strcats onto the sizebuf

	void ClearLink(link_t *l);
	void RemoveLink(link_t *l);
	void InsertLinkBefore(link_t *l, link_t *before);
	void InsertLinkAfter(link_t *l, link_t *after);

	extern	short	(*BigShort)(short l);
	extern	short	(*LittleShort)(short l);
	extern	int		(*BigLong)(int l);
	extern	int		(*LittleLong)(int l);
	extern	float	(*BigFloat)(float l);
	extern	float	(*LittleFloat)(float l);

	void MSG_WriteChar(sizebuf_t *sb, int c);
	void MSG_WriteByte(sizebuf_t *sb, int c);
	void MSG_WriteShort(sizebuf_t *sb, int c);
	void MSG_WriteLong(sizebuf_t *sb, int c);
	void MSG_WriteFloat(sizebuf_t *sb, float f);
	void MSG_WriteString(sizebuf_t *sb, char *s);
	void MSG_WriteCoord(sizebuf_t *sb, float f);
	void MSG_WriteAngle(sizebuf_t *sb, float f);
	void MSG_WriteAngle16(sizebuf_t *sb, float f); //johnfitz

	extern	int			msg_readcount;
	extern	bool	msg_badread;		// set if a read goes beyond end of message

	void MSG_BeginReading();
	int MSG_ReadChar();
	int MSG_ReadByte();
	int MSG_ReadShort();
	int MSG_ReadLong();
	float MSG_ReadFloat();
	char *MSG_ReadString();

	float MSG_ReadCoord();
	float MSG_ReadAngle();
	float MSG_ReadAngle16(); //johnfitz

	void Q_memset(void *dest, int fill, int count);
	void Q_memcpy(void *dest, void *src, int count);
	int Q_memcmp(void *m1, void *m2, int count);
	void Q_strcpy(char *dest, char *src);
	void Q_strncpy(char *dest, char *src, int count);
	int Q_strlen(char *str);
	char *Q_strrchr(char *s, char c);
	void Q_strcat(char *dest, char *src);
	int Q_strcmp(char *s1, char *s2);
	int Q_strncmp(char *s1, char *s2, int count);
	int Q_strcasecmp(char *s1, char *s2);
	int Q_strncasecmp(char *s1, char *s2, int n);
	int	Q_atoi(char *str);
	float Q_atof(char *str);

	extern	char	com_token[1024];
	extern	bool	com_eof;

	char *COM_Parse(char *data);


	extern	int		com_argc;
	extern	char	**com_argv;

	int COM_CheckParm(char *parm);
	void COM_Init(char *path);
	void COM_InitArgv(int argc, char **argv);

	char *COM_SkipPath(char *pathname);
	void COM_StripExtension(char *in, char *out);
	void COM_FileBase(char *in, char *out);
	void COM_DefaultExtension(char *path, char *extension);

	char	*va(char *format, ...);
	// does a varargs printf into a temp buffer

	extern int com_filesize;
	//extern class cache_user_t;

	extern	char	com_gamedir[MAX_OSPATH];

	void COM_WriteFile(char *filename, void *data, int len);
	int COM_OpenFile(char *filename, int *hndl);
	int COM_FOpenFile(char *filename, FILE **file);
	void COM_CloseFile(int h);

	byte *COM_LoadStackFile(char *path, void *buffer, int bufsize);
	byte *COM_LoadTempFile(char *path);
	byte *COM_LoadHunkFile(char *path);
	void COM_LoadCacheFile(char *path, cache_user_t *cu);

	void COM_CreatePath(char *path);

	byte *COM_LoadFile (char *path, int usehunk);
	pack_t *COM_LoadPackFile(char *packfile);
}

extern bool		standard_quake, rogue, hipnotic;

#endif