/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II

	Notes:
	I may or may not have stolen bits from DirectQ ;)
	Thanks mh
======================================================================*/

#include <assert.h>

#include "quakedef.h"
#include "server\server.h"
#include "cmd.h"
#include "console.h"
#include "heap.h"

using namespace Server;

namespace Axiom
{
	cvar_t			*cvar_vars			= NULL;
	cvar_alias_t	*cvar_alias_vars	= NULL;

	char *cvar_null_string = "";

	cvar_t *Cmd_FindCvar(char *name);

	extern int numcomplist;

	extern complist_t *complist;

	/*================================================================================
		Used only for cases where the full completion list can't be
		relied on to be up yet; use Cmd_FindCvar otherwise
	================================================================================*/
	cvar_t *Cvar_FindVar (char *var_name)
	{
		// regular cvars
		for (cvar_t *var = cvar_vars; var; var = var->next)
		{
			if (!_stricmp(var_name, var->name))
			{
				return var;
			}
		}

		// alias cvars
		for(cvar_alias_t *var = cvar_alias_vars; var; var = var->next)
		{
			if (!_stricmp (var_name, var->name))
			{
				return var->var;
			}
		}

		return NULL;
	}

	/*================================================================================
		List
	================================================================================*/
	void Cvar_List_f()
	{
		cvar_t	*cvar;
		char 	*partial;
		int		len, count;

		if (Cmd_Argc() > 1)
		{
			partial	= Cmd_Argv(1);
			len		= strlen(partial);
		}
		else
		{
			partial	= NULL;
			len		= 0;
		}

		count = 0;
		for(cvar = cvar_vars; cvar; cvar = cvar->next)
		{
			if (partial && Q_strncmp (partial,cvar->name, len))
			{
				continue;
			}
			Con_SafePrintf ("%s%s %s \"%s\"\n",
				cvar->usage & CVAR_ARCHIVE ? "*" : " ",
				cvar->usage & CVAR_SERVER ? "s" : " ",
				cvar->name,
				cvar->string);
			count++;
		}

		Con_SafePrintf ("%i cvars", count);
		if (partial)
		{
			Con_SafePrintf (" beginning with \"%s\"", partial);
		}
		Con_SafePrintf ("\n");
	}

	/*================================================================================
		Inc
	================================================================================*/
	void Cvar_Inc_f()
	{
		switch (Cmd_Argc())
		{
		default:
		case 1:
			Con_Printf("inc <cvar> [amount] : increment cvar\n");
			break;
		case 2:
			Cvar_Set(Cmd_Argv(1), Cvar_VariableValue(Cmd_Argv(1)) + 1);
			break;
		case 3:
			Cvar_Set(Cmd_Argv(1), Cvar_VariableValue(Cmd_Argv(1)) + atof(Cmd_Argv(2)));
			break;
		}
	}

	/*================================================================================
		Toggle
	================================================================================*/
	void Cvar_Toggle_f()
	{
		switch (Cmd_Argc())
		{
		default:
		case 1:
			Con_Printf("toggle <cvar> : toggle cvar\n");
			break;
		case 2:
			if (Cvar_VariableValue(Cmd_Argv(1)))
				Cvar_Set (Cmd_Argv(1), "0");
			else
				Cvar_Set (Cmd_Argv(1), "1");
			break;
		}
	}

	/*================================================================================
		Cycle
	================================================================================*/
	void Cvar_Cycle_f()
	{
		int i;

		if (Cmd_Argc() < 3)
		{
			Con_Printf("cycle <cvar> <value list>: cycle cvar through a list of values\n");
			return;
		}

		//loop through the args until you find one that matches the current cvar value.
		//yes, this will get stuck on a list that contains the same value twice.
		//it's not worth dealing with, and i'm not even sure it can be dealt with.

		for(i = 2; i < Cmd_Argc(); i++)
		{
			//zero is assumed to be a string, even though it could actually be zero.  The worst case
			//is that the first time you call this command, it won't match on zero when it should, but after that,
			//it will be comparing strings that all had the same source (the user) so it will work.
			if (atof(Cmd_Argv(i)) == 0)
			{
				if (!strcmp(Cmd_Argv(i), Cvar_VariableString(Cmd_Argv(1))))
					break;
			}
			else
			{
				if (atof(Cmd_Argv(i)) == Cvar_VariableValue(Cmd_Argv(1)))
					break;
			}
		}

		if (i == Cmd_Argc())
			Cvar_Set (Cmd_Argv(1), Cmd_Argv(2)); // no match
		else if (i + 1 == Cmd_Argc())
			Cvar_Set (Cmd_Argv(1), Cmd_Argv(2)); // matched last value in list
		else
			Cvar_Set (Cmd_Argv(1), Cmd_Argv(i+1)); // matched earlier in list
	}

	/*================================================================================
		Init
	================================================================================*/
	void Cvar_Init()
	{
		Cmd_BuildCompletionList();
	}
	
	cmd_t Cvar_List_f_Cmd		("cvarlist",	Cvar_List_f);
	cmd_t Cvar_Toggle_f_Cmd		("toggle",		Cvar_Toggle_f);
	cmd_t Cvar_Cycle_f_Cmd		("cycle",		Cvar_Cycle_f);
	cmd_t Cvar_Inc_f_Cmd		("inc",			Cvar_Inc_f);

	/*================================================================================
		VariableValue
	================================================================================*/
	float Cvar_VariableValue(char *var_name)
	{
		cvar_t	*var;

		var = Cvar_FindVar (var_name);
		if (!var)
			return 0;
		return atof(var->string);
	}

	/*================================================================================
		VariableString
	================================================================================*/
	char *Cvar_VariableString(char *var_name)
	{
		cvar_t *var;

		var = Cvar_FindVar (var_name);
		if (!var)
			return cvar_null_string;

		return var->string;
	}

	/*================================================================================
		SetPrevalidate
	================================================================================*/
	bool Cvar_SetPrevalidate(cvar_t *var)
	{
		if (!var)
		{
			// there is an error in C code if this happens
			Con_Printf ("Cvar_Set: variable %s not found\n", var->name);
			return false;
		}

		// reject set attempt
		if (var->usage & CVAR_READONLY)
		{
			Con_Printf ("Cvar_Set: var->usage & CVAR_READONLY\n");
			return false;
		}

		return true;
	}

	/*================================================================================
		SetBroadcast
	================================================================================*/
	void Cvar_SetBroadcast(cvar_t *var)
	{
		if (!(var->usage & CVAR_SERVER))
			return;

		if (!sv.active)
			return;

		// add the name of the person who changed it to the message
		SV_BroadcastPrintf("\"%s\" was changed to \"%s\" by \"%s\"\n", var->name, var->string, cl_name.string);
	}

	/*================================================================================
		Set
	================================================================================*/
	void Cvar_Set (cvar_t *var, char *value)
	{
		// some QC attempts to set old WinQuake cvars that don't exist any more
		if (!var) return;

		if (!Cvar_SetPrevalidate (var))
			return;

		if (strcmp (var->string, value))
		{
			Zone_Free(var->string);
			var->string		= (char *) Zone_Alloc (strlen (value) + 1);
			strcpy(var->string, value);
			var->value		= atof (var->string);
			var->integer	= (int) var->value;

			Cvar_SetBroadcast (var);
		}
	}

	/*================================================================================
		Set
	================================================================================*/
	void Cvar_Set(cvar_t *var, float value)
	{
		// some QC attempts to set old WinQuake cvars that don't exist any more
		if (!var) return;

		if (!Cvar_SetPrevalidate (var)) return;

		if (var->value != value)
		{
			// store back to the cvar
			var->value = value;
			var->integer = (int) var->value;

			// copy out the value to a temp buffer
			char valbuf[32];
			_snprintf (valbuf, 32, "%g", var->value);

			Zone_Free (var->string);
			var->string = (char *) Zone_Alloc (strlen (valbuf) + 1);

			strcpy (var->string, valbuf);

			Cvar_SetBroadcast (var);
		}
	}

	/*================================================================================
		Set
	================================================================================*/
	void Cvar_Set(char *var_name, float value)
	{
		Cvar_Set(Cmd_FindCvar(var_name), value);
	}

	/*================================================================================
		Set
	================================================================================*/
	void Cvar_Set(char *var_name, char *value)
	{
		Cvar_Set(Cmd_FindCvar(var_name), value);
	}

	/*================================================================================
		Adds a freestanding variable to the variable list.
	================================================================================*/
	static void Cvar_Register(cvar_t *variable)
	{
		// these should never go through here but let's just be certain
		// (edit - actually it does - recursively - when setting up shadows; see below)
		if (variable->usage & CVAR_DUMMY) return;

		// first check to see if it has already been defined
		cvar_t *check = Cvar_FindVar (variable->name);

		// silently ignore it
		if(check)
			return;

		// check for overlap with a command
		if (Cmd_Exists (variable->name)) return;

		// store the value off
		variable->value		= atof(variable->string);
		variable->integer	= (int)variable->value;

		// link the variable in
		variable->next	= cvar_vars;
		cvar_vars		= variable;
	}

	/*================================================================================
		Handles variable inspection and changing from the console
	================================================================================*/
	bool Cvar_Command()
	{
		cvar_t			*v;

		// check variables
		v = Cvar_FindVar (Cmd_Argv(0));
		if (!v)
			return false;

		// perform a variable print or set
		if (Cmd_Argc() == 1)
		{
			Con_Printf ("\"%s\" is \"%s\"\n", v->name, v->string);
			return true;
		}

		Cvar_Set (v->name, Cmd_Argv(1));
		return true;
	}

	/*================================================================================
		Writes lines containing "set variable value" for all variables
		with thr archive flag set to true.
	================================================================================*/
	void Cvar_WriteVariables(FILE *f)
	{
		cvar_t	*var;

		for (var = cvar_vars ; var ; var = var->next)
			if (var->usage & CVAR_ARCHIVE)
				fprintf (f, "%s \"%s\"\n", var->name, var->string);
	}

	/*================================================================================
		cvar_t
	================================================================================*/
	cvar_t::cvar_t (char *cvarname, char *initialval, int useflags)
	{
		// alloc space
		this->name		= (char *)Zone_Alloc (strlen (cvarname) + 1);
		this->string	= (char *)Zone_Alloc (strlen (initialval) + 1);

		// copy in the data
		strcpy (this->name, cvarname);
		strcpy (this->string, initialval);
		this->usage = useflags;

		// self-register the cvar at construction time
		Cvar_Register (this);
	}

	/*================================================================================
		cvar_t
	================================================================================*/
	cvar_t::cvar_t(char *cvarname, float initialval, int useflags)
	{
		// alloc space
		this->name = (char *) Zone_Alloc (strlen (cvarname) + 1);

		// copy out the value to a temp buffer
		char valbuf[32];
		_snprintf(valbuf, 32, "%g", initialval);

		// alloc space for the string
		this->string = (char *) Zone_Alloc (strlen (valbuf) + 1);

		// copy in the data
		strcpy (this->name, cvarname);
		strcpy (this->string, valbuf);
		this->usage = useflags;

		// self-register the cvar at construction time
		Cvar_Register (this);
	}

	/*================================================================================
		cvar_t
	================================================================================*/
	cvar_t::cvar_t()
	{
		// dummy cvar for temp usage; not registered
		this->name		= (char *)Zone_Alloc(2);
		this->string	= (char *)Zone_Alloc(2);

		this->name[0]	= 0;
		this->string[0]	= 0;
		this->value		= 0;
		this->usage		= CVAR_DUMMY;
		this->integer	= 0;
		this->next		= NULL;
	}

	/*================================================================================
		cvar_t
	================================================================================*/
	cvar_t::~cvar_t()
	{
		// protect the zone from overflowing if cvars are declared in function scope
		Zone_Free(this->name);
		Zone_Free(this->string);
	}

	/*================================================================================
		cvar_t
	================================================================================*/
	cvar_alias_t::cvar_alias_t(char *cvarname, cvar_t *cvarvar)
	{
		assert (cvarname);
		assert (cvarvar);

		// these should never go through here but let's just be certain
		// (edit - actually it does - recursively - when setting up shadows; see below)
		if (cvarvar->usage & CVAR_DUMMY) return;

		// first check to see if it has already been defined
		if (Cvar_FindVar (cvarname)) return;

		// check for overlap with a command
		if (Cmd_Exists (cvarname)) return;

		// alloc space for name
		this->name = (char *) Zone_Alloc (strlen (cvarname) + 1);
		strcpy (this->name, cvarname);

		this->var = cvarvar;

		this->next = cvar_alias_vars;
		cvar_alias_vars = this;
	}
}