/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II

	Notes:
	cvar_t variables are used to hold scalar or string variables that can be
	changed or displayed at the console or prog code as well as
	accessed directly in C code.
	
	it is sufficient to initialize a cvar_t with just the first two fields, or
	you can add a ,true flag for variables that you want saved to the configuration
	file when the game is quit:

	cvar_t	r_draworder		= {"r_draworder",	"1"};
	cvar_t	scr_screensize	= {"screensize",	"1", true};

	Cvars must be registered before use, or they will have a 0 value instead of
	the float interpretation of the string.

	Generally, all cvar_t declarations should be registered in the apropriate
	init function before any console commands are executed:
	Cvar_RegisterVariable (&host_framerate);
	
	C code usually just references a cvar in place:
	if ( r_draworder.value )
	
	It could optionally ask for the value to be looked up for a string name:
	if (Cvar_VariableValue ("r_draworder"))

	Interpreted prog code can access cvars with the cvar(name) or
	cvar_set (name, value) internal functions:
	teamplay = cvar("teamplay");
	cvar_set ("registered", "1");

	The user can access cvars from the console in two ways:
	r_draworder			prints the current value
	r_draworder 0		sets the current value to 0
	Cvars are restricted from having the same names as commands to keep this
	interface from being ambiguous.
======================================================================*/

#ifndef CVAR_H
#define CVAR_H

#include <stdio.h>

/*typedef struct cvar_s
{
	char	*name;
	char	*string;
	int		usage;

	//bool	archive;		// set to true to cause it to be saved to vars.rc
	//bool	server;		// notifies players when changed
	float	value;
	struct	cvar_s *next;
	char	*default_string; //johnfitz -- remember defaults for reset function
	void (*callback) (void); //johnfitz
} cvar_t;*/

#define CVAR_SERVER			1
#define CVAR_ARCHIVE		2
#define CVAR_VALUE			4
#define CVAR_DUMMY			8
#define CVAR_HUD			16
#define CVAR_READONLY		32
#define CVAR_COMPAT			128

namespace Axiom
{
	class cvar_t
	{
	public:
		// this one is for use when we just want a dummy cvar (e.g. to take a copy of an existing one for the menus)
		// these types are not registered and do not show up in completion lists
		cvar_t();
		~cvar_t();

		// this one allows specifying of cvars directly by usage flags
		// this is the preferred way and ultimately all cvars will be changed over to this
		cvar_t(char *cvarname, char *initialval, int useflags = 0);

		// same as above but it allows for an explicit value cvar
		cvar_t(char *cvarname, float initialval, int useflags = 0);

		char	*name;
		char	*string;
		float	value;
		int		integer;

		// usage flags
		int usage;

		// next in the chain
		cvar_t *next;
	};

	class cvar_alias_t
	{
	public:
		cvar_alias_t (char *cvarname, cvar_t *cvarvar);
		char *name;
		cvar_t *var;

		cvar_alias_t *next;
	};

	void Cvar_Set(char *var_name, char *value);
	void Cvar_Set(char *var_name, float value);
	void Cvar_Set(cvar_t *var, float value);
	void Cvar_Set(cvar_t *var, char *value);

	float Cvar_VariableValue(char *var_name);

	char *Cvar_VariableString(char *var_name);
	bool Cvar_Command();

	void Cvar_WriteVariables(FILE *f);
	cvar_t *Cvar_FindVar(char *var_name);
	void Cvar_Init();

	extern cvar_t		*cvar_vars;
	extern cvar_alias_t	*cvar_alias_vars;
}

#define Cvar_Get(var, varname) static cvar_t *(var) = NULL; if (!(var)) (var) = Cvar_FindVar (varname);if (!(var)) Sys_Error ("%s is not a cvar", varname);

#endif