/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II

	Notes:
	This whole file is directly stolen from DirectQ
======================================================================*/

#ifndef HEAP_H
#define HEAP_H

extern byte *scratchbuf;
#define SCRATCHBUF_SIZE 262144

// fast memcpy and memset replacements
// memcpy and memset in MS CRT are just straight char-by-char operations!!!
void *Q_MemCpy (void *dst, void *src, int size);
void *Q_MemSet (void *dst, int val, int size);

// interface
void Pool_Init (void);

void *Zone_Alloc (int size);
void Zone_FreeMemory (void *ptr);
void Zone_Compact (void);

// wrapper to ensure that the pointer is NULL after a free op
#define Zone_Free(ptr) {Zone_FreeMemory (ptr); ptr = NULL;}

class CQuakeHunk
{
public:
	CQuakeHunk (int maxsizemb);
	~CQuakeHunk (void);
	void *Alloc (int size);
	void Free (void);

	int GetLowMark (void);
	void FreeToLowMark (int mark);

private:
	void Initialize (void);
	int MaxSize;	// maximum memory reserved by this buffer (converted to bytes in constructor)
	int LowMark;	// current memory pointer position
	int HighMark;	// size of all committed memory so far

	char Name[64];

	byte *BasePtr;
};

class CQuakeZone
{
public:
	CQuakeZone (void);
	~CQuakeZone (void);
	void *Alloc (int size);
	void Free (void *data);
	void Compact (void);
	void Discard (void);

private:
	void EnsureHeap (void);
	HANDLE hHeap;
	int Size;
	int Peak;
};

class CQuakeCache
{
public:
	CQuakeCache (void);
	~CQuakeCache (void);
	void *Alloc (int size);
	void *Alloc (void *data, int size);
	void *Alloc (char *name, void *data, int size);
	void *Check (char *name);
	void Flush (void);

private:
	void Init (void);
	CQuakeZone *Heap;
	struct cacheobject_s *Head;
};

// space buffers
extern CQuakeHunk *MainHunk;
extern CQuakeZone *GameZone;
extern CQuakeZone *MapZone;
extern CQuakeCache *MainCache;
extern CQuakeZone *MainZone;

#endif