#ifndef HOST_H
#define HOST_H

#include "cvar.h"

#define	MINIMUM_MEMORY			0x550000
#define	MINIMUM_MEMORY_LEVELPAK	(MINIMUM_MEMORY + 0x100000)

// stock defines
#define	IT_SHOTGUN				1
#define	IT_SUPER_SHOTGUN		2
#define	IT_NAILGUN				4
#define	IT_SUPER_NAILGUN		8
#define	IT_GRENADE_LAUNCHER		16
#define	IT_ROCKET_LAUNCHER		32
#define	IT_LIGHTNING			64
#define IT_SUPER_LIGHTNING      128
#define IT_SHELLS               256
#define IT_NAILS                512
#define IT_ROCKETS              1024
#define IT_CELLS                2048
#define IT_AXE                  4096
#define IT_ARMOR1               8192
#define IT_ARMOR2               16384
#define IT_ARMOR3               32768
#define IT_SUPERHEALTH          65536
#define IT_KEY1                 131072
#define IT_KEY2                 262144
#define	IT_INVISIBILITY			524288
#define	IT_INVULNERABILITY		1048576
#define	IT_SUIT					2097152
#define	IT_QUAD					4194304
#define IT_SIGIL1               (1<<28)
#define IT_SIGIL2               (1<<29)
#define IT_SIGIL3               (1<<30)
#define IT_SIGIL4               (1<<31)

//rogue changed and added defines
#define RIT_SHELLS              128
#define RIT_NAILS               256
#define RIT_ROCKETS             512
#define RIT_CELLS               1024
#define RIT_AXE                 2048
#define RIT_LAVA_NAILGUN        4096
#define RIT_LAVA_SUPER_NAILGUN  8192
#define RIT_MULTI_GRENADE       16384
#define RIT_MULTI_ROCKET        32768
#define RIT_PLASMA_GUN          65536
#define RIT_ARMOR1              8388608
#define RIT_ARMOR2              16777216
#define RIT_ARMOR3              33554432
#define RIT_LAVA_NAILS          67108864
#define RIT_PLASMA_AMMO         134217728
#define RIT_MULTI_ROCKETS       268435456
#define RIT_SHIELD              536870912
#define RIT_ANTIGRAV            1073741824
#define RIT_SUPERHEALTH         2147483648

//hipnotic added defines
#define HIT_PROXIMITY_GUN_BIT 16
#define HIT_MJOLNIR_BIT       7
#define HIT_LASER_CANNON_BIT  23
#define HIT_PROXIMITY_GUN   (1<<HIT_PROXIMITY_GUN_BIT)
#define HIT_MJOLNIR         (1<<HIT_MJOLNIR_BIT)
#define HIT_LASER_CANNON    (1<<HIT_LASER_CANNON_BIT)
#define HIT_WETSUIT         (1<<(23+2))
#define HIT_EMPATHY_SHIELDS (1<<(23+3))

// the host system specifies the base of the directory tree, the
// command line parms passed to the program, and the amount of memory
// available for the program to use

typedef struct
{
	char	*basedir;
	char	*cachedir;		// for development over ISDN lines
	int		argc;
	char	**argv;
	void	*membase;

	// Feared; int to SIZE_T
	SIZE_T	memsize;
} quakeparms_t;

extern bool noclip_anglehack;

// host
extern	quakeparms_t	host_parms;

extern	cvar_t	sys_ticrate;
extern	cvar_t	sys_nostdout;
extern	cvar_t	developer;
extern	cvar_t	max_edicts; //johnfitz

extern	bool	host_initialized;	// true if into command execution
extern	double	host_frametime;
extern	byte	*host_colormap;
extern	int		host_framecount;	// incremented every frame, never reset
extern	double	realtime;			// not bounded in any way, changed at
									// start of every frame, never reset

namespace Axiom
{
	void Host_ClearMemory();
	void Host_ServerFrame();
	
	void Host_Init(quakeparms_t *parms);
	void Host_Shutdown();
	void Host_Error(char *error, ...);
	void Host_EndGame(char *message, ...);
	void Host_Frame(float time);
	
	void Host_ClientCommands(char *fmt, ...);
	void Host_ShutdownServer(bool crash);

	extern void Host_WriteConfiguration();

	void Host_InitCommands();
	void Host_Quit_f();
}

extern bool		msg_suppress_1;		// suppresses resolution and cache size console output
									//  an fullscreen DIB focus gain/loss
extern int		current_skill;		// skill level for currently loaded level (in case
									//  the user changes the cvar while the level is
									//  running, this reflects the level actually in use)

extern bool		isDedicated;
extern SIZE_T	minimum_memory;

//host_cmd
namespace Axiom
{
	extern void ExtraMaps_Init();
	extern void Modlist_Init();
}

#endif