/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake(props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II
======================================================================*/

#include "pr_comp.h"			// defs shared with qcc
#include "progdefs.h"			// generated by program cdefs

#define	MSG_BROADCAST	0		// unreliable to all
#define	MSG_ONE			1		// reliable to one (msg_entity)
#define	MSG_ALL			2		// reliable to all
#define	MSG_INIT		3		// write to the init string

typedef union eval_s
{
	string_t		string;
	float			_float;
	float			vector[3];
	func_t			function;
	int				_int;
	int				edict;
} eval_t;

#define	MAX_ENT_LEAFS	16
typedef struct edict_s
{
	bool		free;
	link_t			area;				// linked to a division node or leaf
	int				num_leafs;
	short			leafnums[MAX_ENT_LEAFS];
	entity_state_t	baseline;
	unsigned char	alpha;				// johnfitz -- hack to support alpha since it's not part of entvars_t
	bool		sendinterval;		// johnfitz -- send time until nextthink to client for better lerp timing
	float			freetime;			// sv.time when the object was freed
	entvars_t		v;					// C exported fields from progs
// other fields from progs come immediately after
} edict_t;
#define	EDICT_FROM_AREA(l) STRUCT_FROM_LINK(l,edict_t,area)

//============================================================================

extern	dprograms_t		*progs;
extern	dfunction_t		*pr_functions;
extern	char			*pr_strings;
extern	ddef_t			*pr_globaldefs;
extern	ddef_t			*pr_fielddefs;
extern	dstatement_t	*pr_statements;
extern	globalvars_t	*pr_global_struct;
extern	float			*pr_globals;			// same as pr_global_struct

extern	int				pr_edict_size;	// in bytes

//============================================================================

namespace Axiom
{
	namespace Server
	{
		void PR_Init();
		void PR_LoadProgs();

		edict_t *ED_Alloc();
		void ED_Free(edict_t *ed);

		char	*ED_NewString(char *string);
		// returns a copy of the string allocated from the server's string heap

		void ED_Print(edict_t *ed);
		void ED_Write(FILE *f, edict_t *ed);
		char *ED_ParseEdict(char *data, edict_t *ent);

		void ED_WriteGlobals(FILE *f);
		void ED_ParseGlobals(char *data);

		void ED_LoadFromFile(char *data);

		//define EDICT_NUM(n)((edict_t *)(sv.edicts+(n)*pr_edict_size))
		//define NUM_FOR_EDICT(e)(((byte *)(e) - sv.edicts)/pr_edict_size)

		edict_t *EDICT_NUM(int n);
		int NUM_FOR_EDICT(edict_t *e);

		void PR_ExecuteProgram(func_t fnum);
		void PR_Profile_f();
		void PR_RunError(char *error, ...);
	}
}

#define	NEXT_EDICT(e)((edict_t *)((byte *)e + pr_edict_size))

#define	EDICT_TO_PROG(e)((byte *)e -(byte *)sv.edicts)
#define PROG_TO_EDICT(e)((edict_t *)((byte *)sv.edicts + e))

//============================================================================

#define	G_FLOAT(o)(pr_globals[o])
#define	G_INT(o)(*(int *)&pr_globals[o])
#define	G_EDICT(o)((edict_t *)((byte *)sv.edicts+ *(int *)&pr_globals[o]))
#define G_EDICTNUM(o) NUM_FOR_EDICT(G_EDICT(o))
#define	G_VECTOR(o)(&pr_globals[o])
#define	G_STRING(o)(pr_strings + *(string_t *)&pr_globals[o])
#define	G_FUNCTION(o)(*(func_t *)&pr_globals[o])

#define	E_FLOAT(e,o)(((float*)&e->v)[o])
#define	E_INT(e,o)(*(int *)&((float*)&e->v)[o])
#define	E_VECTOR(e,o)(&((float*)&e->v)[o])
#define	E_STRING(e,o)(pr_strings + *(string_t *)&((float*)&e->v)[o])

extern	int		type_size[8];

// Feared; NAMESPACE
namespace Axiom
{
	typedef void(*builtin_t)();
	extern	builtin_t *pr_builtins;
	extern int pr_numbuiltins;

	namespace Server
	{
		void ED_PrintEdicts();
		void ED_PrintNum(int ent);

		eval_t *GetEdictFieldValue(edict_t *ed, char *field);
	}
}

extern int		pr_argc;

extern	bool	pr_trace;
extern	dfunction_t	*pr_xfunction;
extern	int			pr_xstatement;

extern	unsigned short		pr_crc;

namespace Axiom
{
	char *PF_VarString(int first);
	void PF_error();
	void PF_objerror();
	void PF_makevectors();
	void PF_setorigin();
	void SetMinMaxSize(edict_t *e, float *min, float *max, bool rotate);
	void PF_setsize();
	void PF_setmodel();
	void PF_bprint();
	void PF_sprint();
	void PF_centerprint();
	void PF_normalize();
	void PF_vlen();
	void PF_vectoyaw();
	void PF_vectoangles();
	void PF_random();

	void PF_precache_sound(char *s);
}