/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II
======================================================================*/
// disable data conversion warnings

#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA

#ifndef GLQUAKE_H
#define GLQUAKE_H

#ifdef _WIN32
#include <windows.h>
#endif

#include <gl\gl.h>
#include "gl\glext.h"
#include "gl\wglext.h"

#include "gl_extlist.h"

#include "client\client.h"
#include "cvar.h"
#include "renderer\gl_texmgr.h"
#include "bspfile.h"
#include "vid.h"
#include "mathlib.h"

void VID_LockBuffer();
void VID_UnlockBuffer();

void GL_BeginRendering (int *x, int *y, int *width, int *height);
void GL_EndRendering ();

//johnfitz -- removed texture object stuff since they are standard in gl 1.1

typedef struct
{
	float	x, y, z;
	float	s, t;
	float	r, g, b;
} glvert_t;

extern glvert_t glv;

extern	int glx, gly, glwidth, glheight;

#ifdef _WIN32
extern	PROC glArrayElementEXT;
extern	PROC glColorPointerEXT;
extern	PROC glTexturePointerEXT;
extern	PROC glVertexPointerEXT;
#endif

// r_local.h -- private refresh defs

#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
					// normalizing factor so player model works out to about
					//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		480

#define TILE_SIZE		128		// size of textures generated by R_GenTiledSurf

#define SKYSHIFT		7
#define	SKYSIZE			(1 << SKYSHIFT)
#define SKYMASK			(SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01


void R_TimeRefresh_f ();

namespace Axiom
{
	void R_ReadPointFile_f();
	texture_t *R_TextureAnimation(texture_t *base, int frame);
}

typedef struct surfcache_s
{
	struct surfcache_s	*next;
	struct surfcache_s 	**owner;		// NULL is an empty chunk of memory
	int					lightadj[MAXLIGHTMAPS]; // checked for strobe flush
	int					dlight;
	int					size;		// including header
	unsigned			width;
	unsigned			height;		// DEBUG only needed for debug
	float				mipscale;
	struct texture_s	*texture;	// checked for animating textures
	byte				data[4];	// width*height elements
} surfcache_t;


typedef struct
{
	pixel_t		*surfdat;	// destination for generated surface
	int			rowbytes;	// destination logical width in bytes
	msurface_t	*surf;		// description for surface to generate
	fixed8_t	lightadj[MAXLIGHTMAPS];
							// adjust for lightmap levels for dynamic lighting
	texture_t	*texture;	// corrected for animating textures
	int			surfmip;	// mipmapped ratio of surface texels / world pixels
	int			surfwidth;	// in mipmapped texels
	int			surfheight;	// in mipmapped texels
} drawsurf_t;


typedef enum {
	pt_static, pt_grav, pt_slowgrav, pt_fire, pt_explode, pt_explode2, pt_blob, pt_blob2
} ptype_t;

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
typedef struct particle_s
{
// driver-usable fields
	vec3_t		org;
	float		color;
// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;
	ptype_t		type;
} particle_t;


//====================================================

extern	bool	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	entity_t	*currententity;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;
extern	mplane_t	frustum[4];

//
// view origin
//
extern	vec3_t	vup;
extern	vec3_t	vpn;
extern	vec3_t	vright;
extern	vec3_t	r_origin;

//
// screen size info
//
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	int		d_lightstylevalue[256];	// 8.8 fraction of base light value

extern	bool	envmap;

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_speeds;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_fullbright;
extern	cvar_t	r_lightmap;
extern	cvar_t	r_shadows;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;

extern	cvar_t	gl_clear;
extern	cvar_t	gl_cull;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_flashblend;
extern	cvar_t	gl_nocolors;

extern	cvar_t	gl_max_size;
extern	cvar_t	gl_playermip;

extern	float	r_world_matrix[16];

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;

void R_TranslatePlayerSkin (int playernum);
void R_TranslateNewPlayerSkin (int playernum); //johnfitz -- this handles cases when the actual texture changes

#ifndef _WIN32
#define APIENTRY /* */
#endif

//johnfitz -- anisotropic filtering
#define	GL_TEXTURE_MAX_ANISOTROPY_EXT 0x84FE
#define	GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT 0x84FF
//johnfitz

//johnfitz -- polygon offset
#define OFFSET_BMODEL 1
#define OFFSET_NONE 0
#define OFFSET_DECAL -1
#define OFFSET_FOG -2
#define OFFSET_SHOWTRIS -3
void GL_PolygonOffset (int);
//johnfitz

//johnfitz -- GL_EXT_texture_env_combine
//the values for GL_ARB_ are identical
#define GL_COMBINE_EXT			0x8570
#define GL_COMBINE_RGB_EXT		0x8571
#define GL_COMBINE_ALPHA_EXT	0x8572
#define GL_RGB_SCALE_EXT		0x8573
#define GL_CONSTANT_EXT			0x8576
#define GL_PRIMARY_COLOR_EXT	0x8577
#define GL_PREVIOUS_EXT			0x8578
#define GL_SOURCE0_RGB_EXT		0x8580
#define GL_SOURCE1_RGB_EXT		0x8581
#define GL_SOURCE0_ALPHA_EXT	0x8588
#define GL_SOURCE1_ALPHA_EXT	0x8589
extern bool gl_texture_env_combine;
//johnfitz

extern bool gl_texture_env_add; //johnfitz -- for GL_EXT_texture_env_add

extern bool isIntelVideo; //johnfitz -- intel video workarounds from Baker

//johnfitz -- rendering statistics
extern int rs_brushpolys, rs_aliaspolys, rs_skypolys, rs_particles, rs_fogpolys;
extern int rs_dynamiclightmaps, rs_brushpasses, rs_aliaspasses, rs_skypasses;
extern float rs_megatexels;
//johnfitz

//johnfitz -- track developer statistics that vary every frame
extern cvar_t devstats;
//typedef struct devstats_t
typedef struct
{
	int		packetsize;
	int		edicts;
	int		visedicts;
	int		efrags;
	int		tempents;
	int		beams;
	int		dlights;
}  devstats_t;

extern devstats_t dev_stats;
extern devstats_t dev_peakstats;

//johnfitz

//ohnfitz -- reduce overflow warning spam
typedef struct
{
	double	packetsize;
	double	efrags;
	double	beams;
} overflowtimes_t;

extern overflowtimes_t dev_overflows; //this stores the last time overflow messages were displayed, not the last time overflows occured

#define CONSOLE_RESPAM_TIME 3 // seconds between repeated warning messages
//johnfitz

//johnfitz -- moved here from r_brush.c
#define MAX_LIGHTMAPS 256 //johnfitz -- was 64
extern gltexture_t *lightmap_textures[MAX_LIGHTMAPS]; //johnfitz -- changed to an array
//johnfitz

extern bool r_drawflat_cheatsafe, r_fullbright_cheatsafe, r_lightmap_cheatsafe, r_drawworld_cheatsafe; //johnfitz

// gl_fog.cpp
namespace Axiom
{
	void Fog_ParseServerMessage ();
	float *Fog_GetColor();
	float Fog_GetDensity();
	void Fog_EnableGFog();
	void Fog_DisableGFog();
	void Fog_StartAdditive();
	void Fog_StopAdditive();
	void Fog_SetupFrame();
	void Fog_NewMap();
	void Fog_Init();

	float Fog_GetDensity();
	float *Fog_GetColor();
}

// rlight
extern void R_AnimateLight();
extern void R_RenderDlights();
extern int R_LightPoint(vec3_t p);
extern void R_MarkLights(dlight_t *light, int bit, mnode_t *node);

// gl_sky
extern void Sky_LoadSkyBox(char *name);
extern void Sky_LoadTexture(texture_t *mt);
extern void Sky_DrawSky();
extern void Sky_Init();
extern void Sky_NewMap();

// warp
extern void GL_SubdivideSurface(msurface_t *fa);
extern void DrawWaterPoly(glpoly_t *p);
extern int gl_warpimagesize; //johnfitz -- for water warp

// mesh
namespace Axiom
{
	extern void GL_MakeAliasModelDisplayLists(model_t *, aliashdr_t *);
}

// misc
extern void R_NewGame();

// gl_refrag
extern void R_StoreEfrags(efrag_t **ppefrag);

// gl_screen
extern void SCR_LoadPics();

// brush
namespace Axiom
{
	extern void R_DrawBrushModel(entity_t *e);
	extern void R_DrawBrushModel_ShowTris(entity_t *e);
	extern void R_RebuildAllLightmaps();
	extern void GL_BuildLightmaps();
	extern void DrawGLPoly(glpoly_t *p);
	extern void R_RenderDynamicLightmaps(msurface_t *fa);
	extern void DrawGLTriangleFan(glpoly_t *p);
	extern void R_UploadLightmap(int lmap);

	// sprite
	extern void R_DrawSpriteModel(entity_t *e);
}

// part
namespace Axiom
{
	extern void R_DrawParticles_ShowTris();
	extern void R_DrawParticles();
	extern void R_InitParticles();
	extern void R_ClearParticles();
	extern void CL_RunParticles();
}

// model
extern void Mod_SetExtraFlags(model_t *mod);

// main
extern bool R_CullModelForEntity(entity_t *e);
extern void R_RotateForEntity(vec3_t origin, vec3_t angles);
extern bool R_CullBox(vec3_t emins, vec3_t emaxs);

// texmgr
extern void TexMgr_RecalcWarpImageSize();

#endif
