/*======================================================================
	See GNU.txt, CREDITS.txt, and COPYING.txt
	Quake owned by ID Software, Inc.
	Code based on FitzQuake (props to John Fitzgibbons)
	QuakeCPP by Brian "Feared" Wychopen II
======================================================================*/

#ifndef WORLD_H
#define WORLD_H

#include "progs.h"

typedef struct
{
	vec3_t	normal;
	float	dist;
} plane_t;

typedef struct
{
	bool	allsolid;	// if true, plane is not valid
	bool	startsolid;	// if true, the initial point was in a solid area
	bool	inopen, inwater;
	float	fraction;		// time completed, 1.0 = didn't hit anything
	vec3_t	endpos;			// final position
	plane_t	plane;			// surface normal at impact
	edict_t	*ent;			// entity the surface is on
} trace_t;

#define	MOVE_NORMAL		0
#define	MOVE_NOMONSTERS	1
#define	MOVE_MISSILE	2

namespace Axiom
{
	void SV_ClearWorld (void);
	void SV_UnlinkEdict (edict_t *ent);
	void SV_LinkEdict (edict_t *ent, bool touch_triggers);
	int SV_PointContents (vec3_t p);
	int SV_TruePointContents (vec3_t p);
	edict_t	*SV_TestEntityPosition (edict_t *ent);

	trace_t SV_Move (vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end, int type, edict_t *passedict);
}

extern void R_UpdateWarpTextures();

namespace Axiom
{
	extern bool SV_RecursiveHullCheck(hull_t *hull, int num, float p1f, float p2f, vec3_t p1, vec3_t p2, trace_t *trace);

	extern void R_MarkSurfaces();
	extern void R_CullSurfaces();
	extern void R_DrawTextureChains_ShowTris();
	extern void R_DrawWorld();
	extern void R_DrawTextureChains_Water();
}

#endif